library reflector.mirrors;

import 'dart:mirrors';
import 'package:reflector/src/interfaces.dart';

/// Runtime implementation of [Reflector] that uses dart:mirrors.
class MirrorsReflector implements Reflector {
  final Iterable<DeclarationMirror> _mirrors;

  /// Returns a [MirrorReflector] that is scoped to reflect on [types]. All
  /// provided [types] must be classes, or an [ArgumentError] is thrown.
  factory MirrorsReflector.fromClasses(List<Type> types) {
    return new MirrorsReflector._(types.map((type) {
      return reflectClass(type);
    }));
  }

  const MirrorsReflector._(this._mirrors);

  @override
  Iterable<RuntimeMetadata> call([Query query = const Query()]) {
    return _mirrors.where((mirror) {
      // Apply filter: query.withAnnotations.
      if (query.withAnnotations.isNotEmpty) {
        final annotationTypes = mirror.metadata.map((m) {
          return m.type.reflectedType;
        }).toSet();

        if (!annotationTypes.containsAll(query.withAnnotations)) {
          return false;
        }
      }

      // No filters failed.
      return true;
    }).map((mirror) {
      if (mirror is ClassMirror) {
        // Inclusion: query.includeAnnotations.
        var annotations = const [];
        if (query.includeAnnotations) {
          annotations = mirror.metadata.map((m) {
            return m.reflectee;
          }).toList(growable: false);
        }

        return new ClassRuntimeMetadata(
            (mirror as ClassMirror).reflectedType,
            mirror.simpleName,
            mirror.qualifiedName,
            annotations: annotations);
      } else {
        throw new UnimplementedError('Not implemented: ${mirror.runtimeType}.');
      }
    });
  }
}
